import torch
import math
from .ActivationFunction import ActivationFunction


class CaLU(ActivationFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = torch.arctan(x) / math.pi
        b = 0.5
        return x * (a + b)



import torch
from .SoftCriticModule import SoftCriticModule


class DualSoftCriticModule(torch.nn.Module):
    def __init__(
        self, state_dimension: int, hidden_dimension: int, action_dimension: int
    ):
        super().__init__()
        self.alpha = SoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )
        self.beta = SoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )

    def forward(self, state: torch.Tensor, action: torch.Tensor):
        return self.alpha(state, action), self.beta(state, action)



import torch
import math
from .ActivationFunction import ActivationFunction


class ReSP(ActivationFunction):
    def __init__(self, alpha: float = 1.5):
        super().__init__()
        self.alpha = alpha

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.where(
            x >= 0, self.alpha * x + math.log(2), torch.log(1 + torch.exp(x))
        )



import torch
import numpy
from ..Dataset import Dataset


class CoraDataset(Dataset):
    def __init__(self):
        pass

    def load(path: str):
        idx_features_labels = numpy.genfromtxt(f"{path}/cora.content", dtype=str)
        features = torch.tensor(idx_features_labels[:, 1:-1].astype(numpy.float32))
        classes = {label: i for i, label in enumerate(set(idx_features_labels[:, -1]))}
        labels = torch.tensor(
            [classes[label] for label in idx_features_labels[:, -1]], dtype=torch.long
        )

        # map node ids to index
        ids = idx_features_labels[:, 0].astype(numpy.int32)
        id_map = {j: i for i, j in enumerate(ids)}

        # read edges
        edges_unordered = numpy.genfromtxt(f"{path}/cora.cites", dtype=numpy.int32)
        edges = numpy.array(
            list(map(lambda x: [id_map[x[0]], id_map[x[1]]], edges_unordered))
        )

        # build adjacency matrix
        n = labels.shape[0]
        adj = torch.zeros((n, n))
        for i, j in edges:
            adj[i, j] = 1
            adj[j, i] = 1

        # normalize features
        features = features / features.sum(1, keepdim=True).clamp(min=1)

        # simple splits
        idx_train = torch.arange(140)
        idx_val = torch.arange(200, 500)
        idx_test = torch.arange(500, 1500)

        return features, adj, labels, idx_train, idx_val, idx_test



import fire
from .namespaces import MakeNamespace, CacheNamespace, MediaNamespace
from .commands import InitCommand


class HomaCommand:
    def __init__(self):
        self.make = MakeNamespace()
        self.cache = CacheNamespace()
        self.media = MediaNamespace()

    def init(self):
        InitCommand.run()


def main():
    fire.Fire(HomaCommand)



import numpy
import random
import torch
from types import SimpleNamespace
from .Buffer import Buffer
from ...device import get_device


class SoftActorCriticBuffer(Buffer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def record(
        self,
        state: numpy.ndarray,
        action: numpy.ndarray,
        reward: float,
        next_state: numpy.ndarray,
        termination: float,
    ):
        self.collection.append((state, action, reward, next_state, termination))

    def sample(self, k: int, as_tensor: bool = False, move_to_device: bool = True):
        batch = random.sample(self.collection, k)
        states, actions, rewards, next_states, terminations = zip(*batch)

        states = numpy.array(states, dtype=numpy.float32)
        actions = numpy.array(actions, dtype=numpy.float32)
        rewards = numpy.array(rewards, dtype=numpy.float32)
        next_states = numpy.array(next_states, dtype=numpy.float32)
        terminations = numpy.array(terminations, dtype=numpy.float32)

        # add one dimension to both rewards and terminations
        rewards = numpy.expand_dims(rewards, axis=-1)
        terminations = numpy.expand_dims(terminations, axis=-1)

        if as_tensor:
            states = torch.from_numpy(states).float()
            actions = torch.from_numpy(actions).float()
            rewards = torch.from_numpy(rewards).float()
            next_states = torch.from_numpy(next_states).float()
            terminations = torch.from_numpy(terminations).float()

        if as_tensor and move_to_device:
            _device = get_device()
            states = states.to(_device)
            actions = actions.to(_device)
            rewards = rewards.to(_device)
            next_states = next_states.to(_device)
            terminations = terminations.to(_device)

        return SimpleNamespace(
            **{
                "states": states,
                "actions": actions,
                "rewards": rewards,
                "next_states": next_states,
                "terminations": terminations,
            }
        )

    def sample_torch(self, k: int):
        return self.sample(k=k, as_tensor=True)



import torch
from .ActivationFunction import ActivationFunction


class Serf(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        return x * torch.erf(torch.log(1 + torch.exp(x)))



import torch
from .MultiHeadGraphAttentionModule import MultiHeadGraphAttentionModule


class GraphAttentionModule(torch.nn.Module):
    def __init__(
        self,
        input_dimension: int,
        hidden_dimension: int,
        output_dimension: int,
        num_heads: int,
        dropout: float,
        alpha: float,
        concat: bool,
        activation: torch.nn.Module,
        final_activation: torch.nn.Module,
        v2: bool,
    ):
        super().__init__()
        self.theta = MultiHeadGraphAttentionModule(
            input_dimension=input_dimension,
            output_dimension=hidden_dimension,
            num_heads=num_heads,
            dropout=dropout,
            alpha=alpha,
            concat=concat,
            activation=activation,
            final_activation=final_activation,
            v2=v2,
        )
        self.sigma = MultiHeadGraphAttentionModule(
            input_dimension=hidden_dimension * num_heads,
            output_dimension=output_dimension,
            num_heads=1,
            dropout=dropout,
            alpha=alpha,
            concat=concat,
            activation=activation,
            final_activation=final_activation,
            v2=v2,
        )
        self.dropout = torch.nn.Dropout(dropout)

    def forward(self, features: torch.Tensor, adjacency_matrix: torch.Tensor):
        features = self.dropout(features)
        features = self.theta(features, adjacency_matrix)
        features = self.dropout(features)
        features = self.sigma(features, adjacency_matrix)
        return torch.nn.functional.log_softmax(features, dim=1)



import torch
from .diayn.Actor import Actor
from .diayn.Critic import Critic
from .diayn.Discriminator import Discriminator
from .buffers import DiversityIsAllYouNeedBuffer, Buffer


class DiversityIsAllYouNeed:
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        hidden_dimension: int = 256,
        num_skills: int = 10,
        critic_decay: float = 0.0,
        actor_decay: float = 0.0,
        discriminator_decay: float = 0.0,
        actor_lr: float = 0.0001,
        critic_lr: float = 0.001,
        discriminator_lr=0.001,
        buffer_capacity: int = 1_000_000,
        actor_epsilon: float = 1e-6,
        gamma: float = 0.99,
        min_std: float = -20.0,
        max_std: float = 2.0,
    ):
        self.buffer: Buffer = DiversityIsAllYouNeedBuffer(capacity=buffer_capacity)
        self.num_skills: int = num_skills
        self.actor = Actor(
            state_dimension=state_dimension,
            action_dimension=action_dimension,
            hidden_dimension=hidden_dimension,
            num_skills=num_skills,
            lr=actor_lr,
            decay=actor_decay,
            epsilon=actor_epsilon,
            min_std=min_std,
            max_std=max_std,
        )
        self.critic = Critic(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            num_skills=num_skills,
            lr=critic_lr,
            decay=critic_decay,
            gamma=gamma,
        )
        self.discriminator = Discriminator(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            num_skills=num_skills,
            lr=discriminator_lr,
            decay=discriminator_decay,
        )

    def one_hot(self, indices, max_index) -> torch.Tensor:
        one_hot = torch.zeros(indices.size(0), max_index)
        one_hot.scatter_(1, indices.unsqueeze(1), 1)
        return one_hot

    def skill_index(self) -> torch.Tensor:
        return torch.randint(0, self.num_skills, (1,))

    def skill(self) -> torch.Tensor:
        return self.one_hot(self.skill_index(), self.num_skills)

    def advantages(
        self,
        states: torch.Tensor,
        skills: torch.Tensor,
        rewards: torch.Tensor,
        terminations: torch.Tensor,
        next_states: torch.Tensor,
    ) -> torch.Tensor:
        values = self.critic.values(states=states, skills=skills)
        termination_mask = 1 - terminations
        next_values = self.critic.values_(states=next_states, skills=skills)
        update = self.gamma * next_values * termination_mask
        return rewards + update - values

    def train(self, skill: torch.Tensor):
        data = self.buffer.all_tensor()
        skill_indices = skill.repeat(data.states.size(0), 1).long()
        skills_indices_one_hot = self.one_hot(skill_indices, self.num_skills)
        self.discriminator.train(
            states=data.states, skills_indices=skills_indices_one_hot
        )
        advantages = self.advantages(
            states=data.states,
            rewards=data.rewards,
            terminations=data.terminations,
            next_states=data.next_states,
            skills=skills,
        )
        self.critic.train(advantages=advantages)
        self.actor.train(advantages=advantages)



import torch
import numpy
from .modules import DiscriminatorModule
from ...common.concerns import MovesModulesToDevice


class Discriminator(MovesModulesToDevice):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        num_skills: int,
        decay: float,
        lr: float,
    ):
        self.num_skills: int = num_skills

        self.network = DiscriminatorModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            num_skills=num_skills,
        )
        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=decay
        )
        self.criterion = torch.nn.CrossEntropyLoss()
        self.move_modules()

    def loss(self, states: torch.Tensor, skills_indices: torch.Tensor):
        logits = self.network(states)
        return self.criterion(logits, skills_indices)

    @torch.no_grad()
    def reward(self, state: torch.Tensor, skill_index: torch.Tensor):
        logits = self.network(state)
        probabilities = torch.nn.functional.log_softmax(logits, dim=-1)
        entropy = numpy.log(1.0 / self.num_skills)
        if skill_index.dim() == 1:
            skill_index = skill_index.unsqueeze(-1)
        reward = probabilities.gather(1, skill_index.long()) - entropy
        return reward.squeeze(-1)

    def train(self, states: torch.Tensor, skills_indices: torch.Tensor):
        self.optimizer.zero_grad()
        loss = self.loss(states=states, skills_indices=skills_indices)
        loss.backward()
        self.optimizer.step()



from __future__ import annotations

import torch
from typing import TYPE_CHECKING
from torch.nn.functional import mse_loss as mse
from .modules import SoftCriticModule
from ..utils import soft_update
from ...common.concerns import MovesModulesToDevice

if TYPE_CHECKING:
    from .SoftActor import SoftActor


class SoftCritic(MovesModulesToDevice):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        action_dimension: int,
        lr: float,
        weight_decay: float,
        gamma: float,
    ):
        self.gamma: float = gamma

        self.zeta = SoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )
        self.eta = SoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )
        self.zeta_target = SoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )
        self.eta_target = SoftCriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
        )

        # copy source to target when initiated
        self.zeta_target.load_state_dict(self.zeta.state_dict())
        self.eta_target.load_state_dict(self.eta.state_dict())

        self.move_modules()

        self.optimizer = torch.optim.AdamW(
            list(self.zeta.parameters()) + list(self.eta.parameters()),
            lr=lr,
            weight_decay=weight_decay,
        )

    def train(
        self,
        states: torch.Tensor,
        actions: torch.Tensor,
        rewards: torch.Tensor,
        terminations: torch.Tensor,
        next_states: torch.Tensor,
        actor: SoftActor,
        alpha: float,
    ):
        self.zeta.train()
        self.eta.train()

        self.optimizer.zero_grad()
        loss = self.loss(
            states=states,
            actions=actions,
            rewards=rewards,
            terminations=terminations,
            next_states=next_states,
            actor=actor,
            alpha=alpha,
        )
        loss.backward()
        self.optimizer.step()

    def loss(
        self,
        states: torch.Tensor,
        actions: torch.Tensor,
        rewards: torch.Tensor,
        terminations: torch.Tensor,
        next_states: torch.Tensor,
        actor: torch.nn.Module,
        alpha: float,
    ):
        q_zeta = self.zeta(states, actions)
        q_eta = self.eta(states, actions)
        target = self.calculate_target(
            rewards=rewards,
            terminations=terminations,
            next_states=next_states,
            actor=actor,
            alpha=alpha,
        )
        return mse(q_zeta, target) + mse(q_eta, target)

    @torch.no_grad()
    def calculate_target(
        self,
        rewards: torch.Tensor,
        terminations: torch.Tensor,
        next_states: torch.Tensor,
        actor: SoftActor,
        alpha: float,
    ):
        termination_mask = 1 - terminations
        next_actions, next_probabilities = actor.sample(next_states)
        q_zeta_target = self.zeta_target(next_states, next_actions)
        q_eta_target = self.eta_target(next_states, next_actions)
        q_target = torch.min(q_zeta_target, q_eta_target)
        entropy_q = q_target - alpha * next_probabilities
        return rewards + self.gamma * termination_mask * entropy_q

    def update(self, tau: float):
        soft_update(network=self.zeta, target=self.zeta_target, tau=tau)
        soft_update(network=self.eta, target=self.eta_target, tau=tau)



import os
import sys
import ffmpeg
from .Namespace import Namespace


class MediaNamespace(Namespace):
    def _validate_input(self, input_filename: str) -> bool:
        if not os.path.exists(input_filename):
            print(f"Error: Input file not found at {input_filename}", file=sys.stderr)
            return False
        return True

    def _get_output_filename(self, input_filename: str, extension: str) -> str:
        base, _ = os.path.splitext(input_filename)
        return f"{base}.{extension}"

    def _run_ffmpeg(self, input_filename: str, output_filename: str, **output_kwargs):
        try:
            (
                ffmpeg.input(input_filename)
                .output(output_filename, **output_kwargs)
                .run(capture_stdout=True, capture_stderr=True, overwrite_output=True)
            )
            print(f"✅ Successfully converted and saved to {output_filename}")
        except ffmpeg.Error as e:
            print("FFmpeg Error:", file=sys.stderr)
            print(e.stderr.decode(), file=sys.stderr)
        except Exception as e:
            print(f"Unexpected error: {e}", file=sys.stderr)

    def _convert_to_webm(self, input_filename: str):
        if not self._validate_input(input_filename):
            return
        output_filename = self._get_output_filename(input_filename, "webm")
        self._run_ffmpeg(
            input_filename,
            output_filename,
            vcodec="libvpx-vp9",
            crf=32,
            acodec="libopus",
            audio_bitrate="96k",
        )

    def _convert_to_gif(self, input_filename: str):
        if not self._validate_input(input_filename):
            return
        output_filename = self._get_output_filename(input_filename, "gif")

        fps = 10
        scale_width = 400
        max_colors = 64
        dither_mode = "bayer"

        complex_filter = (
            f"fps={fps},scale={scale_width}:-1:flags=lanczos,split [a][b]; "
            f"[a] palettegen=stats_mode=diff:max_colors={max_colors} [p]; "
            f"[b][p] paletteuse=dither={dither_mode}:diff_mode=rectangle"
        )

        self._run_ffmpeg(
            input_filename,
            output_filename,
            filter_complex=complex_filter,
            loop=0,
        )

    def convert(self, target_format: str, *input_files):
        convert_map = {
            "web": self._convert_to_webm,
            "webm": self._convert_to_webm,
            "gif": self._convert_to_gif,
        }

        handler = convert_map.get(target_format.lower())
        if not handler:
            print(
                f"Error: Unsupported target format '{target_format}'", file=sys.stderr
            )
            return

        for input_file in input_files:
            handler(input_file)



import torch
from .SoftActorCriticRepository import SoftActorCriticRepository
from ...common.concerns import MovesModulesToDevice


class SoftActorCriticTemperature(MovesModulesToDevice):
    def __init__(
        self,
        lr: float,
        weight_decay: float,
        action_dimension: int,
        device: torch.device,
    ):
        self.log_alpha = torch.nn.Parameter(torch.zeros(1, device=device))
        self.optimizer = torch.optim.Adam(
            [self.log_alpha], lr=lr, weight_decay=weight_decay
        )
        self.target_entropy = -float(action_dimension)

    @property
    def alpha(self):
        return self.log_alpha.exp()

    def train(self):
        self.optimizer.zero_grad()
        loss = self.loss()
        loss.backward()
        self.optimizer.step()

    def loss(self) -> torch.Tensor:
        # updated inside actor
        log_probability = SoftActorCriticRepository.log_probability

        theta = (-log_probability - self.target_entropy).detach()
        return (self.alpha * theta).mean()



import torch
from typing import Type


class DiscriminatorModule(torch.nn.Module):
    def __init__(self, state_dimension: int, hidden_dimension: int, num_skills: int):
        super().__init__()
        self.state_dimension: int = state_dimension
        self.hidden_dimension: int = hidden_dimension
        self.num_skills: int = num_skills

        self.phi: Type[torch.nn.Sequential] = torch.nn.Sequential(
            torch.nn.Linear(self.state_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hidden_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
        )
        self.fc: Type[torch.nn.Linear] = torch.nn.Linear(
            self.hidden_dimension, self.num_skills
        )

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        features: torch.Tensor = self.phi(state)
        return self.fc(features)



import torch
from .ActivationFunction import ActivationFunction


class PoLU(ActivationFunction):
    def __init__(self, alpha: float = 1.5):
        super().__init__()
        if alpha not in [1, 1.5, 2]:
            raise ValueError(f"PoLU alpha must be in [1, 2, 1.5]. Got {alpha}.")

        self.alpha = alpha

    def forward(self, x: torch.Tensor):
        delta = 1 - x
        return torch.where(x >= 0, x, delta.pow(-self.alpha) - 1)



import torch
from .ActivationFunction import ActivationFunction


class SoftModulusQ(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.where(torch.abs(x) <= 1, x.pow(2) * (2 - x.abs()), x.abs())



import torch
from .ActivationFunction import ActivationFunction


class ASiLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        alpha = 1 / (1 + torch.exp(-x))
        return torch.arctan(x * alpha)



class ReportsEnsembleSize:
    def __init__(self):
        super().__init__()

    @property
    def size(self):
        return len(self.weights)

    @property
    def length(self):
        return self.size



from .ShiftedReLU import ShiftedReLU
from .ADA import ADA
from .OAF import OAF
from .AbsLU import AbsLU
from .ParametricLogish import ParametricLogish
from .ExpExpish import ExpExpish
from .DoubleSiLU import DoubleSiLU
from .GeneralizedSwish import GeneralizedSwish
from .MSiLU import MSiLU
from .TBSReLU import TBSReLU
from .ASiLU import ASiLU
from .NoisyReLU import NoisyReLU
from .ExponentialDLReLU import ExponentialDLReLU
from .SaRa import SaRa
from .SiELU import SiELU
from .EANAF import EANAF
from .MaxSig import MaxSig
from .TangentSigmoidReLU import TangentSigmoidReLU
from .Phish import Phish
from .SelfArctan import SelfArctan
from .PFLU import PFLU
from .ReSP import ReSP
from .Serf import Serf
from .LogSigmoid import LogSigmoid
from .SlopedReLU import SlopedReLU
from .ReCU import ReCU
from .MinSin import MinSin
from .LaLU import LaLU
from .mReLU import mReLU
from .FlattedTSwish import FlattedTSwish
from .ERF import ERF
from .RePU import RePU
from .TangentBipolarSigmoidReLU import TangentBipolarSigmoidReLU
from .BaseDLReLU import BaseDLReLU
from .Logish import Logish
from .TripleStateSwish import TripleStateSwish
from .ReQU import ReQU
from .ExponentialSwish import ExponentialSwish
from .SinSig import SinSig
from .PLAF import PLAF
from .TeLU import TeLU
from .DiffELU import DiffELU
from .Elliot import Elliot
from .SoftModulusQ import SoftModulusQ
from .DerivativeSiLU import DerivativeSiLU
from .NLReLU import NLReLU
from .IpLU import IpLU
from .ThLU import ThLU
from .ActivationFunction import ActivationFunction
from .RReLU import RReLU
from .PolyLU import PolyLU
from .Suish import Suish
from .TSiLU import TSiLU
from .SoftsignRReLU import SoftsignRReLU
from .SoftModulusT import SoftModulusT
from .Gish import Gish
from .DRLU import DRLU
from .NReLU import NReLU
from .LogLogish import LogLogish
from .SGELU import SGELU
from .TSReLU import TSReLU
from .REU import REU
from .ReSech import ReSech
from .SineReLU import SineReLU
from .DLReLU import DLReLU
from .DLU import DLU
from .CaLU import CaLU
from .RandomizedSlopedReLU import RandomizedSlopedReLU
from .TanhExp import TanhExp
from .PoLU import PoLU
from .GCU import GCU
from .SigmoidDerivative import SigmoidDerivative
from .Smish import Smish
from .SigLU import SigLU
from .AdaptiveActivationFunction import AdaptiveActivationFunction



from .sac import SoftActor
from .tqc import QuantileMultiCritic


class TQC:
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        hidden_dimension: int = 256,
        actor_lr: float = 0.0003,
        actor_decay: float = 0.0,
        alpha: float = 0.2,
        min_std: float = -20.0,
        max_std: float = 2.0,
        num_quantiles: int = 25,
        num_critics: int = 5,
    ):
        self.critics = QuantileMultiCritic(
            num_critics=num_critics, num_quantiles=num_quantiles
        )
        self.actor = SoftActor(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
            lr=actor_lr,
            weight_decay=actor_decay,
            alpha=alpha,
            min_std=min_std,
            max_std=max_std,
        )



import torch


class QuantileCriticModule(torch.nn.Module):
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        hidden_dimension: int,
        num_quantiles: int,
    ):
        super().__init__()
        self.phi = torch.nn.Sequential(
            torch.nn.Linear(state_dimension + action_dimension, hidden_dimension),
            torch.nn.ReLU(),
            torch.nn.Linear(hidden_dimension, hidden_dimension),
            torch.nn.ReLU(),
        )
        self.fc = torch.nn.Linear(hidden_dimension, num_quantiles)

    def forward(self, state: torch.Tensor, action: torch.Tensor) -> torch.Tensor:
        psi = torch.cat([state, action], dim=-1)
        features = self.phi(psi)
        return self.fc(features)



from .CalculatesMetricNecessities import CalculatesMetricNecessities
from .PredictsProbabilities import PredictsProbabilities
from .ReportsClassificationMetrics import ReportsClassificationMetrics
from .ReportsEnsembleAccuracy import ReportsEnsembleAccuracy
from .ReportsEnsembleF1 import ReportsEnsembleF1
from .ReportsEnsembleKappa import ReportsEnsembleKappa
from .ReportsLogits import ReportsLogits
from .ReportsEnsembleSize import ReportsEnsembleSize
from .StoresModels import StoresModels
from .SavesEnsembleModels import SavesEnsembleModels



import torch
from .ActivationFunction import ActivationFunction


class ReSech(ActivationFunction):
    def __init__(self):
        super().__init__()

    def sech(self, x: torch.Tensor) -> torch.Tensor:
        return torch.tensor(2) / (torch.exp(x) + torch.exp(-x))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.sech(x) * x



import torch
import random


def replace_activations(module, needle: torch.nn.Module, candidates: list):
    for name, child in module.named_children():
        if isinstance(child, needle):
            new_activation = random.choice(candidates)
            setattr(module, name, new_activation())
        else:
            replace_activations(child, needle, candidates)
    return module



import torch
from ....common.modules import EncoderModule


class SoftCriticModule(torch.nn.Module):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        action_dimension: int,
    ):
        super().__init__()

        self.phi = EncoderModule(
            input_dimension=state_dimension, hidden_dimension=hidden_dimension
        )
        self.chi = EncoderModule(
            input_dimension=action_dimension + hidden_dimension,
            hidden_dimension=hidden_dimension,
        )
        self.fc = torch.nn.Linear(hidden_dimension, 1)

    def forward(self, state: torch.Tensor, action: torch.Tensor):
        state = self.phi(state)
        features = torch.cat([state, action], dim=1)
        features = self.chi(features)
        return self.fc(features)



import torch
from typing import List, Type
from collections import OrderedDict
from ...vision import Model


class StoresModels:
    def __init__(self):
        super().__init__()
        self.factories: List[Type[torch.nn.Module]] = []
        self.weights: List[OrderedDict] = []

    def record(self, model: Model | torch.nn.Module):
        model_: torch.nn.Module | None = None
        if isinstance(model, Model):
            model_ = model.network
        elif isinstance(model, torch.nn.Module):
            model_ = model
        else:
            raise TypeError("Wrong input to ensemble record")

        self.factories.append(model_.__class__)
        self.weights.append(model_.state_dict())

    def push(self, *args, **kwargs):
        self.record(*args, **kwargs)

    def append(self, *args, **kwargs):
        self.record(*args, **kwargs)

    def add(self, *args, **kwargs):
        self.record(*args, **kwargs)



import torch
from .ActivationFunction import ActivationFunction


class SaRa(ActivationFunction):
    def __init__(self):
        super().__init__()
        self.alpha = 0.5
        self.beta = 0.7

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        delta = 1 + self.alpha * torch.exp(-self.beta * x)
        return torch.where(x >= 0, x, x / delta)



import shutil
import sys
from pathlib import Path
from .Namespace import Namespace


class CacheNamespace(Namespace):
    def clear(self):
        root = Path.cwd()
        removed = 0
        errors: list[str] = []

        for candidate in root.rglob("__pycache__"):
            if not candidate.is_dir():
                continue
            try:
                shutil.rmtree(candidate)
                removed += 1
            except OSError as exc:
                errors.append(f"{candidate}: {exc}")

        if errors:
            print("Failed to remove the following paths:", file=sys.stderr)
            for error in errors:
                print(f"  - {error}", file=sys.stderr)
            return 1

        return (
            f"Removed {removed} __pycache__ director{'ies' if removed != 1 else 'y'}."
        )



from .concerns import (
    ReportsEnsembleSize,
    StoresModels,
    ReportsClassificationMetrics,
    PredictsProbabilities,
    SavesEnsembleModels,
)


class Ensemble(
    ReportsEnsembleSize,
    ReportsClassificationMetrics,
    PredictsProbabilities,
    StoresModels,
    SavesEnsembleModels,
):
    def __init__(self):
        super().__init__()



import torch
from sklearn.metrics import f1_score, cohen_kappa_score
from types import SimpleNamespace
from typing import Type, OrderedDict
from .modules import GraphAttentionModule
from ..common.concerns import MovesModulesToDevice


class GraphAttention(MovesModulesToDevice):
    def __init__(
        self,
        labels: torch.Tensor,
        features: torch.Tensor,
        adjacency_matrix: torch.Tensor,
        input_dimension: int,
        output_dimension: int,
        hidden_dimension: int = 8,
        num_heads: int = 8,
        alpha: float = 0.2,
        lr: float = 0.005,
        decay: float = 0.0005,
        dropout: float = 0.6,
        concat: bool = True,
        activation: torch.nn.Module = torch.nn.LeakyReLU,
        final_activation: torch.nn.Module = torch.nn.ELU,
        v2: bool = False,
    ):
        super().__init__()

        self.features = features
        self.adjacency_matrix = adjacency_matrix
        self.labels = labels

        self.network = GraphAttentionModule(
            input_dimension=input_dimension,
            output_dimension=output_dimension,
            hidden_dimension=hidden_dimension,
            num_heads=num_heads,
            dropout=dropout,
            alpha=alpha,
            concat=concat,
            activation=activation,
            final_activation=final_activation,
            v2=v2,
        )
        self.move_modules()

        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=decay
        )

        # this is used because model outputs log-probability distribution
        self.criterion = torch.nn.NLLLoss()

    def train(self, mask: torch.Tensor) -> None:
        self.network.train()
        self.optimizer.zero_grad()
        loss = self.loss(mask=mask)
        loss.backward()
        self.optimizer.step()

    def loss(self, mask: torch.Tensor) -> torch.Tensor:
        predictions = self.network(self.features, self.adjacency_matrix)
        masked_predictions = predictions[mask]
        masked_labels = self.labels[mask]
        return self.criterion(masked_predictions, masked_labels)

    def state_dict(self):
        return self.network.state_dict()

    def load_state_dict(self, state_dict: OrderedDict):
        self.network.load_state_dict(state_dict)

    @torch.no_grad()
    def metrics(self, mask: torch.Tensor) -> Type[SimpleNamespace]:
        self.network.eval()

        predictions = self.network(self.features, self.adjacency_matrix)
        masked_predictions = predictions[mask].argmax(dim=1).cpu()
        masked_labels = self.labels[mask].cpu()

        accuracy = (masked_predictions == masked_labels).float().mean().item()
        f1_macro = f1_score(masked_labels, masked_predictions, average="micro")
        kappa = cohen_kappa_score(masked_labels, masked_predictions)

        return SimpleNamespace(
            **{
                "accuracy": accuracy,
                "f1": f1_macro,
                "kappa": kappa,
            }
        )



import torch
from .ActivationFunction import ActivationFunction


class GeneralizedSwish(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return x * torch.sigmoid(torch.exp(-x))



import torch
from .ActivationFunction import ActivationFunction


class MaxSig(ActivationFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: torch.Tensor):
        return torch.max(x, torch.sigmoid(x))



import torch
from ...device import get_device


class MovesModulesToDevice:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def move_modules(self):
        for module in dir(self):
            if module.startswith("__") or module.endswith("__"):
                continue

            attribute = getattr(self, module)
            _device = get_device()
            if (
                isinstance(attribute, torch.nn.Module)
                or isinstance(attribute, torch.nn.Parameter)
                or isinstance(attribute, torch.Tensor)
            ):
                if hasattr(attribute, "to"):
                    try:
                        setattr(self, module, attribute.to(_device))
                    except (AttributeError, TypeError):
                        # it passes properties
                        pass



import torch
import numpy
from types import SimpleNamespace
from .Buffer import Buffer
from .concerns import HasRecordAlternatives


class DiversityIsAllYouNeedBuffer(Buffer, HasRecordAlternatives):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def all_tensor(self) -> SimpleNamespace:
        return self.all(tensor=True)

    def all(self, tensor: bool = False) -> SimpleNamespace:
        states, actions, rewards, next_states, terminations, probabilities = zip(
            *self.collection
        )

        if tensor:
            states = torch.from_numpy(numpy.array(states))
            actions = torch.from_numpy(numpy.array(actions))
            rewards = torch.from_numpy(numpy.array(rewards))
            next_states = torch.from_numpy(numpy.array(next_states))
            terminations = torch.from_numpy(numpy.array(terminations))
            probabilities = torch.from_numpy(numpy.array(probabilities))

        return SimpleNamespace(
            **{
                "states": states,
                "actions": actions,
                "rewards": rewards,
                "next_states": next_states,
                "terminations": terminations,
                "probabilities": probabilities,
            }
        )

    def record(
        self,
        state: numpy.ndarray,
        action: int,
        reward: float,
        next_state: numpy.ndarray,
        termination: bool,
        probability: numpy.ndarray,
    ) -> None:
        self.collection.append(
            (state, action, reward, next_state, termination, probability)
        )



import torch
from .modules import CriticModule
from ...common.concerns import MovesModulesToDevice


class Critic(MovesModulesToDevice):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        num_skills: int,
        lr: float,
        decay: float,
        gamma: float,
    ):
        self.gamma: float = gamma

        self.network = CriticModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            num_skills=num_skills,
        )
        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=decay
        )
        self.criterion = torch.nn.SmoothL1Loss()
        self.move_modules()

    def train(self, advantages: torch.Tensor):
        self.optimizer.zero_grad()
        loss = self.loss(advantages=advantages)
        loss.backward()
        self.optimizer.step()

    def loss(self, advantages: torch.Tensor):
        return advantages.pow(2).mean()

    def values(self, states: torch.Tensor, skills: torch.Tensor):
        return self.network(states, skills)

    @torch.no_grad()
    def values_(self, *args, **kwargs):
        return self.values(*args, **kwargs)



from __future__ import annotations

import numpy
import torch
from typing import TYPE_CHECKING
from .modules import SoftActorModule
from .SoftActorCriticRepository import SoftActorCriticRepository
from ...common.concerns import MovesModulesToDevice
from ...device import get_device

if TYPE_CHECKING:
    from .SoftCritic import SoftCritic


class SoftActor(MovesModulesToDevice):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        action_dimension: int,
        lr: float,
        weight_decay: float,
        min_std: float,
        max_std: float,
    ):
        self.network = SoftActorModule(
            state_dimension=state_dimension,
            hidden_dimension=hidden_dimension,
            action_dimension=action_dimension,
            min_std=min_std,
            max_std=max_std,
        )
        self.move_modules()

        self.optimizer = torch.optim.AdamW(
            self.network.parameters(), lr=lr, weight_decay=weight_decay
        )

    def train(self, states: torch.Tensor, critic: SoftCritic, alpha: float):
        self.network.train()
        self.optimizer.zero_grad()
        loss = self.loss(states=states, critic=critic, alpha=alpha)
        loss.backward()
        self.optimizer.step()

    def loss(
        self, states: torch.Tensor, critic: SoftCritic, alpha: float
    ) -> torch.Tensor:
        actions, log_probability = self.sample(states)

        # used to train temperature
        SoftActorCriticRepository.log_probability = log_probability

        q_zeta = critic.zeta(states, actions)
        q_eta = critic.eta(states, actions)
        q = torch.min(q_zeta, q_eta)
        return (alpha * log_probability - q).mean()

    def process_state(
        self, state: numpy.ndarray | torch.Tensor, move_to_device: bool = True
    ) -> torch.Tensor:
        if isinstance(state, numpy.ndarray):
            state = torch.from_numpy(state).float()

        if state.ndim < 2:
            state = state.unsqueeze(0)

        if move_to_device:
            state = state.to(get_device())

        return state

    def sample(self, state: numpy.ndarray | torch.Tensor):
        state = self.process_state(state)

        mean, log_std = self.network(state)
        std = log_std.exp()

        distribution = torch.distributions.Normal(mean, std)
        z = distribution.rsample()
        action = torch.tanh(z)
        log_probability = distribution.log_prob(z) - torch.log(1 - action.pow(2) + 1e-6)
        log_probability = log_probability.sum(dim=1, keepdim=True)
        return action, log_probability



import torch
from .ActivationFunction import ActivationFunction


class REU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.where(x >= 0, x, x * torch.exp(x))



import torch
from .ActivationFunction import ActivationFunction


class SoftModulusT(ActivationFunction):
    def __init__(self, alpha: float = 0.01):
        super().__init__()
        self.alpha = alpha

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.tanh(x / self.alpha)



import torch
from .ActivationFunction import ActivationFunction


class BaseDLReLU(ActivationFunction):
    def __init__(self, a: float = 0.01, init_mse: float = 1.0, mode: str = "linear"):
        super().__init__()
        assert 0.0 < a < 1.0, "a must be in (0,1)"
        assert mode in ("linear", "exp")
        self.a = float(a)
        self.mode = mode
        self.register_buffer("prev_mse", torch.tensor(float(init_mse)))

    @torch.no_grad()
    def set_prev_mse(self, mse_value):
        if isinstance(mse_value, torch.Tensor):
            mse_value = float(mse_value.detach().cpu().item())
        self.prev_mse.fill_(mse_value)

    @torch.no_grad()
    def update_from_loss(self, loss_tensor: torch.Tensor):
        self.set_prev_mse(loss_tensor)

    def forward(self, z: torch.Tensor, mse_prev: torch.Tensor | float | None = None):
        b_t = (
            self.prev_mse
            if mse_prev is None
            else (torch.as_tensor(mse_prev, device=z.device, dtype=z.dtype))
        )
        if self.mode == "linear":
            slope = self.a * b_t
        else:
            slope = self.a * torch.exp(-b_t)
        return torch.where(z >= 0, z, slope * z)



import torch
from .ActivationFunction import ActivationFunction


class SlopedReLU(ActivationFunction):
    def __init__(self, alpha: float = 5.0):
        super().__init__()
        if not (1 <= alpha < 10):
            raise Exception("Sloped ReLU slope needs to be in [1, 10].")
        self.alpha: float = alpha

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.where(x >= 0, self.alpha * x, 0)



import torch


class GraphAttentionHeadModule(torch.nn.Module):
    def __init__(
        self,
        input_dimension: int,
        output_dimension: int,
        dropout: float,
        alpha: float,
        activation: torch.nn.Module,
        final_activation: torch.nn.Module,
        v2: bool,
    ):
        super().__init__()
        self.v2: bool = v2

        self.phi = torch.nn.Linear(input_dimension, output_dimension, bias=True)

        # separate source, target attentions
        self.mu = torch.nn.Linear(output_dimension, 1, bias=True)
        self.xi = torch.nn.Linear(output_dimension, 1, bias=True)

        # LeakyReLU is a special case because in the original paper, they have explicitly defined
        # alpha = 0.2
        if activation is torch.nn.LeakyReLU:
            self.activation = torch.nn.LeakyReLU(alpha)
        else:
            self.activation = activation()

        self.final_activation = final_activation()
        self.dropout = torch.nn.AlphaDropout(dropout)
        self.norm = torch.nn.LayerNorm(output_dimension)

    def forward(self, features: torch.Tensor, adj: torch.Tensor):
        # computes features
        h = self.phi(features)
        h = self.norm(h)

        # computes attention coefficients
        if not self.v2:
            e = self.mu(h) + self.xi(h).T
            e = self.activation(e)
        else:
            scores = h.unsqueeze(1) + h.unsqueeze(0)
            e = self.mu(scores).squeeze(-1)
            e = self.activation(e)

        e = e.masked_fill(adj == 0, float("-inf"))

        a = torch.nn.functional.softmax(e, dim=1)
        a = self.dropout(a)
        return self.final_activation(a @ h)



import torch


class ContinuousActorModule(torch.nn.Module):
    def __init__(
        self,
        state_dimension: int,
        action_dimension: int,
        hidden_dimension: int,
        num_skills: int,
        min_std: float,
        max_std: float,
    ):
        super().__init__()
        self.state_dimension: int = state_dimension
        self.action_dimension: int = action_dimension
        self.num_skills: int = num_skills
        self.hidden_dimension: int = hidden_dimension
        self.input_dimension: int = self.state_dimension + self.num_skills
        self.min_std: float = min_std
        self.max_std: float = max_std

        self.phi = torch.nn.Sequential(
            torch.nn.Linear(self.input_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hidden_dimension, self.hidden_dimension),
            torch.nn.ReLU(),
            torch.nn.Linear(self.hidden_dimension, self.hidden_dimension),
        )

        self.mu = torch.nn.Linear(self.hidden_dimension, self.action_dimension)
        self.xi = torch.nn.Linear(self.hidden_dimension, self.action_dimension)

    def forward(self, state: torch.Tensor, skill: torch.Tensor) -> torch.Tensor:
        # fix the size to be one state per batch
        state = state.view(state.size(0), -1)

        psi = torch.cat([state, skill], dim=-1)
        features = self.phi(psi)
        mean = self.mu(features)
        std = self.xi(features).clamp(self.min_std, self.max_std)
        return mean, std



import torch
from ...AdaptiveActivationFunction import AdaptiveActivationFunction
from ...concerns import ChannelBased


class LeLeLU(AdaptiveActivationFunction, ChannelBased):
    def __init__(self):
        super().__init__()
        self.a = None

    def forward(self, x: torch.Tensor):
        self.initialize(x, "a")
        a = self.a.view(self.parameter_shape(x))
        return torch.where(x >= 0, a * x, 0.01 * a * x)



import torch
from .ActivationFunction import ActivationFunction


class PFLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        delta = x / (torch.sqrt(1 + x.pow(2)))
        return x * 0.5 * (1 + delta)



import torch
from .GraphAttentionHeadModule import GraphAttentionHeadModule


class MultiHeadGraphAttentionModule(torch.nn.Module):
    def __init__(
        self,
        input_dimension: int,
        output_dimension: int,
        num_heads: int,
        dropout: float,
        alpha: float,
        concat: bool,
        activation: torch.nn.Module,
        final_activation: torch.nn.Module,
        v2: bool,
    ):
        super().__init__()

        self.num_heads: int = num_heads
        self.output_dimension: int = output_dimension
        self.concat: bool = concat

        self.heads = torch.nn.ModuleList(
            [
                GraphAttentionHeadModule(
                    input_dimension=input_dimension,
                    output_dimension=output_dimension,
                    dropout=dropout,
                    alpha=alpha,
                    activation=activation,
                    final_activation=final_activation,
                    v2=v2,
                )
                for _ in range(num_heads)
            ]
        )

    def forward(self, features: torch.Tensor, adjacency_matrix: torch.Tensor):
        features = [head(features, adjacency_matrix) for head in self.heads]

        # aggregate results either by concatenating them or averaging over
        if self.concat:
            features = torch.cat(features, dim=1)
        else:
            features = torch.mean(torch.stack(features), dim=0)

        return features



import torch
from ...AdaptiveActivationFunction import AdaptiveActivationFunction
from ...concerns import ChannelBased


class FReLU(AdaptiveActivationFunction, ChannelBased):
    def __init__(self):
        super().__init__()
        self.b = None

    def forward(self, x: torch.Tensor):
        self.initialize(x, "b")
        b = self.b.view(self.parameter_shape(x))
        return torch.where(x >= 0, x + b, b)



import torch
from .ActivationFunction import ActivationFunction


class SiELU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        chi = x + 0.044715 * x.pow(3)
        delta = 2 * torch.sqrt(torch.tensor(2) / torch.pi) * chi
        return x * torch.sigmoid(delta)



from torch import Tensor
from torch.utils.data.dataloader import DataLoader
from .HasLogits import HasLogits
from .HasProbabilities import HasProbabilities
from .HasLabels import HasLabels
from ...device import get_device


class Trainable(HasLogits, HasProbabilities, HasLabels):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def train(self, x: Tensor | DataLoader, y: Tensor | None = None):
        if y is None and isinstance(x, DataLoader):
            self.train_dataloader(x)
            return
        self.train_tensors(x, y)

    def train_tensors(self, x: Tensor, y: Tensor):
        self.network.train()
        self.optimizer.zero_grad()
        loss = self.criterion(self.network(x).float(), y)
        loss.backward()
        self.optimizer.step()

    def train_dataloader(self, dataloader: DataLoader):
        for x, y in dataloader:
            x, y = x.to(get_device()), y.to(get_device())
            self.train_tensors(x, y)



import torch


class ReportsLogits:
    def __init__(self):
        super().__init__()

    def logits_average(self, x: torch.Tensor) -> torch.Tensor:
        return self.logits_sim(x) / len(self.factories)

    def logits_sum(self, x: torch.Tensor) -> torch.Tensor:
        batch_size = x.size(0)
        logits = torch.zeros((batch_size, self.num_classes))
        for factory, weight in zip(self.factories, self.weights):
            model = factory(num_classes=self.num_classes)
            model.load_state_dict(weight)
            logits += model(x)
        return logits

    def check_aggregation_strategy(self, aggregation: str):
        if aggregation not in ["mean", "average", "sum"]:
            raise ValueError(
                f"Ensemble aggregation strategy must be in [mean, average, sum], but found {aggregation}."
            )

    def logits(self, x: torch.Tensor, aggregation: str = "mean") -> torch.Tensor:
        self.check_aggregation_strategy(aggregation=aggregation)
        logits_handlers = {
            "mean": self.logits_average,
            "average": self.logits_average,
            "sum": self.logits_sum,
        }
        handler = logits_handlers.get(aggregation)
        return handler(x)

    @torch.no_grad()
    def logits_(self, *args, **kwargs):
        return self.logits(*args, **kwargs)



import torch


class LogLogish(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        a = -torch.exp(x)
        return x * (1 - torch.exp(a))



import torch
from .ActivationFunction import ActivationFunction


class Suish(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = x * torch.exp(-torch.abs(x))
        return torch.max(x, a)



import torch
from .ActivationFunction import ActivationFunction


class PolyLU(ActivationFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x: torch.Tensor):
        return torch.where(x >= 0, x, 1 / (1 - x) - 1)



from collections import deque


class ResetsCollection:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def reset(self):
        self.collection = deque(maxlen=self.capacity)



import torch
from .ActivationFunction import ActivationFunction


class DoubleSiLU(ActivationFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = 1 + torch.exp(-x)
        b = -x * 1 / a
        c = 1 + torch.exp(b)
        return x * 1 / c



import torch


def get_model_device(model: torch.nn.Module):
    try:
        device = next(model.parameters()).device
    except StopIteration:
        device = torch.device("cpu")
    return device



import torch
from ...AdaptiveActivationFunction import AdaptiveActivationFunction
from ...concerns import ChannelBased


class DualLine(AdaptiveActivationFunction, ChannelBased):
    def __init__(self):
        super().__init__()
        self.a = None
        self.b = None
        self.m = None

    def forward(self, x: torch.Tensor):
        self.initialize(x, ["a", "b", "m"], [1, 0.01, -0.22])
        a = self.a.view(self.parameter_shape(x))
        b = self.b.view(self.parameter_shape(x))
        m = self.m.view(self.parameter_shape(x))
        return torch.where(x >= 0, a * x + m, b * x + m)



import torch
from ....common.modules import EncoderModule


class SoftActorModule(torch.nn.Module):
    def __init__(
        self,
        state_dimension: int,
        hidden_dimension: int,
        action_dimension: int,
        min_std: float,
        max_std: float,
    ):
        super().__init__()

        self.state_dimension: int = state_dimension
        self.hidden_dimension: int = hidden_dimension
        self.action_dimension: int = action_dimension
        self.min_std: float = float(min_std)
        self.max_std: float = float(max_std)
        self.std_difference: float = self.max_std - self.min_std

        self.phi = EncoderModule(
            input_dimension=state_dimension, hidden_dimension=hidden_dimension
        )

        self.mu = torch.nn.Linear(self.hidden_dimension, self.action_dimension)
        torch.nn.init.uniform_(self.mu.weight, -3e-3, 3e-3)
        torch.nn.init.constant_(self.mu.bias, 0.0)

        self.xi = torch.nn.Linear(self.hidden_dimension, self.action_dimension)
        torch.nn.init.constant_(self.xi.weight, 0.0)
        torch.nn.init.constant_(self.xi.bias, -0.5)

    def forward(self, state: torch.Tensor):
        features = self.phi(state)
        mean = self.mu(features)
        log_std = self.xi(features)
        log_std = self.min_std + (log_std + 1) * 0.5 * self.std_difference
        return mean, log_std
